/*
 * Copyright 2005 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"$Id: nss-emulation.c,v 1.2 2006/03/02 06:31:36 njacobs Exp $"

#include <stdio.h>
#include <pthread.h>
#include <dlfcn.h>
#include <errno.h>
#include <string.h>
#include <nss-emulation.h>

/*
 * This module does a VERY rough emulation of the functionality provided
 * by the name service switch.  It specifically supports the "printers"
 * database and doesn't support the action symantics.  Also, the backend
 * routines must be linked into the library.  It doesn't dlopen() a module
 * and then locate the function like the switch does.
 */

#define NSSWITCH	"/etc/nsswitch.conf"
#define DB		"printers"

#ifndef RTLD_DEFAULT
#define RTLD_DEFAULT 0
#endif

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
static char **source_list = NULL;


static char **
get_source_list()
{
	if (source_list == NULL) {
		FILE *fp;

		pthread_mutex_lock(&lock);
		if (source_list != NULL) {
			/* it was filled in while we waited */
			pthread_mutex_unlock(&lock);
			return (source_list);
		}

		if ((fp = fopen(NSSWITCH, "r")) != NULL) {
			char line[BUFSIZ];

			while (fgets(line, sizeof (line), fp) != NULL) {
				char *tmp, *iter = NULL;

				tmp = strtok_r(line, ":\t\n ", &iter);
				if (tmp == NULL)
					continue;
				if (strcmp(tmp, DB) == 0) {
					while ((tmp = strtok_r(NULL, ":\t\n ",
							&iter)) != NULL)
						list_append(&source_list,
							strdup(tmp));
					break;
				}
			}
			fclose(fp);
		}

		if (source_list == NULL) {
			static char *_list[] = {"files", "printcap",
						"nis", NULL};
			source_list = _list;
		}
		pthread_mutex_unlock(&lock);
	}

	return (source_list);
}

void *
find_func(char *ns, char *func)
{
	char name[64];
	void *result = NULL;

	snprintf(name, sizeof (name), "_nss_%s_%s", ns, func);
	if ((result = dlsym(RTLD_DEFAULT, name)) == NULL) {
		void *handle;
		char lib[BUFSIZ];

		snprintf(lib, sizeof (lib), "/usr/lib/libnss%s.so", ns);
		if (((handle = dlopen(lib, RTLD_NOW)) != NULL) &&
		    ((result = dlsym(handle, name)) == NULL))
			dlclose(handle);
	}

	return (result);
}

void
emul_setprinterentry(int stayopen)
{
	char **list = NULL;

	if ((list = get_source_list()) != NULL) {
		while (*list != NULL) {
			void (*fpt)(int);
			
			if ((fpt = (void (*)(int))find_func(*list, "setprinterentry")) != NULL)
                		(*fpt)(stayopen);
			list++;
		}
	}
}

void
emul_endprinterentry()
{
	char **list = NULL;

	if ((list = get_source_list()) != NULL) {
		while (*list != NULL) {
			void (*fpt)();
			
			if ((fpt = (void (*)(int))find_func(*list, "endprinterentry")) != NULL)
                		(*fpt)();
			list++;
		}
	}
}

int
emul_getprinterentry_r(char *buffer, size_t buflen)
{
	nss_status_t rc = NSS_UNAVAIL;
	char **list = NULL;

	if ((list = get_source_list()) != NULL) {
		while ((*list != NULL) && (rc != NSS_SUCCESS)) {
			nss_status_t (*fpt)(char *, size_t, int *);
			int err = 0;
			
			if ((fpt = (nss_status_t (*)(char *, size_t))find_func(*list, "getprinterentry_r")) != NULL)
                		rc = (*fpt)(buffer, buflen, &err);
			list++;
		}
	}
	if (rc != NSS_SUCCESS)
		buffer[0] = '\0';

	return ((int)rc);
}

int
emul_getprinterbyname_r(const char *name, char *buffer, size_t buflen)
{
	nss_status_t rc = NSS_UNAVAIL;
	char **list = NULL;

	if ((list = get_source_list()) != NULL) {
		while ((*list != NULL) && (rc != NSS_SUCCESS)) {
			nss_status_t (*fpt)(const char *, char *, size_t);
			
			if ((fpt = (nss_status_t (*)(const char *, char *, size_t))find_func(*list, "getprinterbyname_r")) != NULL)
                		rc = (*fpt)(name, buffer, buflen);
			list++;
		}
	}
	if (rc != NSS_SUCCESS)
		buffer[0] = '\0';

	return ((int)rc);
}
