/*
 * Copyright 2005 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/* 
 * Copyright (C) 2001 by Free Software Foundation, Inc.
 * All Rights Reserved
 */

/* This file is part of the GNU C Library.
   Contributed by Norm Jacobs <Norm.Jacobs@sun.com>

   This was based on the Mail alias file parser in nss_files module.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#pragma ident	"$Id: nss-files.c,v 1.1 2005/02/26 06:58:42 njacobs Exp $"

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <pthread.h>

#include "nss-emulation.h"


/* Locks the static variables in this file.  */
static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

/* Maintenance of the shared stream open on the database file.  */

static FILE *stream;
static fpos_t position;
static enum { none, getent, getby } last_use;


static int
_read_line(FILE *f, char *buffer, int buflen)
{
	int linelen;
	int c;

	while (1) {
		linelen = 0;
		while (linelen < (buflen - 1)) {
			switch (c = getc_unlocked(f)) {
			case EOF:
				if (linelen == 0 ||
				    buffer[linelen - 1] == '\\') {
					return (-1);
				} else {
					buffer[linelen    ] = '\n';
					buffer[linelen + 1] = '\0';
					return (linelen);
				}
			case '\n':
				if (linelen > 0 &&
				    buffer[linelen - 1] == '\\') {
					--linelen;  /* remove the '\\' */
				} else {
					buffer[linelen    ] = '\n';
					buffer[linelen + 1] = '\0';
					return (linelen);
				}
				break;
			default:
				buffer[linelen++] = c;
			}
		}
		/* Buffer overflow -- eat rest of line and loop again */
		/* ===> Should syslog() */
		do {
			c = getc_unlocked(f);
			if (c == EOF) {
				return (-1);
			}
		} while (c != '\n');
	}
	/*NOTREACHED*/
}


static nss_status_t
internal_setent (void)
{
  nss_status_t status = NSS_SUCCESS;

  if (stream == NULL)
    {
      stream = fopen ("/etc/printers.conf", "r");

      if (stream == NULL)
	status = errno == EAGAIN ? NSS_TRYAGAIN : NSS_UNAVAIL;
      else
	{
	  /* We have to make sure the file is  `closed on exec'.  */
	  int result, flags;

	  result = flags = fcntl (fileno (stream), F_GETFD, 0);
	  if (result >= 0)
	    {
	      flags |= FD_CLOEXEC;
	      result = fcntl (fileno (stream), F_SETFD, flags);
	    }
	  if (result < 0)
	    {
	      /* Something went wrong.  Close the stream and return a
		 failure.  */
	      fclose (stream);
	      stream = NULL;
	      status = NSS_UNAVAIL;
	    }
	}
    }
  else
    rewind (stream);

  return status;
}


/* Thread-safe, exported version of that.  */
nss_status_t
_nss_files_setprinterentry (void)
{
  nss_status_t status;

  pthread_mutex_lock(&lock);

  status = internal_setent ();

  if (status == NSS_SUCCESS && fgetpos (stream, &position) < 0)
    {
      fclose (stream);
      stream = NULL;
      status = NSS_UNAVAIL;
    }

  last_use = getent;

  pthread_mutex_unlock(&lock);

  return status;
}


/* Close the database file.  */
static void
internal_endent (void)
{
  if (stream != NULL)
    {
      fclose (stream);
      stream = NULL;
    }
}


/* Thread-safe, exported version of that.  */
nss_status_t
_nss_files_endprintentry (void)
{
  pthread_mutex_lock(&lock);

  internal_endent ();

  pthread_mutex_unlock(&lock);

  return NSS_SUCCESS;
}


static nss_status_t
get_next_printer (const char *match, char *buffer, size_t buflen, int *errnop)
{
  nss_status_t status = NSS_NOTFOUND;

  while (1) {
    char *p = buffer;

    memset(buffer, 0, buflen);
    if (_read_line(stream, buffer, buflen) < 0)
		return status;

    while (isspace (*p))
		p++;

    if ((*p == '#')	|| (*p == '\0')) /* the line is a comment */
		continue;

    if (match != NULL) {
		char *pos = buffer;

    		status = NSS_RETURN;

		while (((pos = strstr(pos, match)) != NULL) &&
		       (status != NSS_SUCCESS)) {
			char next = pos[strlen(match)];

			if ((next == '|') || (next == ':'))
				status = NSS_SUCCESS;
			pos++;
		}
    } else
        status = NSS_SUCCESS;
    break;
  }

  return status;
}


nss_status_t
_nss_files_getprinterentry_r (char *buffer, size_t buflen, int *errnop)
{
  /* Return next entry in host file.  */
  nss_status_t status = NSS_SUCCESS;

  pthread_mutex_lock(&lock);

  /* Be prepared that the set*ent function was not called before.  */
  if (stream == NULL)
    status = internal_setent ();

  if (status == NSS_SUCCESS)
    {
      /* If the last use was not by the getent function we need the
	 position the stream.  */
      if (last_use != getent)
	{
	  if (fsetpos (stream, &position) < 0)
	    status = NSS_UNAVAIL;
	  else
	    last_use = getent;
	}

      if (status == NSS_SUCCESS)
	{
	  /* Read lines until we get a definite result.  */
	  do {
	    status = get_next_printer (NULL, buffer, buflen, errnop);
	  } while (status == NSS_RETURN);

	  /* If we successfully read an entry remember this position.  */
	  if (status == NSS_SUCCESS)
	    fgetpos (stream, &position);
	  else
	    last_use = none;
	}
    }

  pthread_mutex_unlock(&lock);

  return status;
}


nss_status_t
_nss_files_getprinterbyname_r (const char *name, 
			     char *buffer, size_t buflen, int *errnop)
{
  /* Return next entry in host file.  */
  nss_status_t status = NSS_SUCCESS;

  if (name == NULL)
    {
      /* __set_errno (EINVAL); */
      errno = EINVAL;
      return NSS_UNAVAIL;
    }

  pthread_mutex_lock(&lock);

  /* Open the stream or rest it.  */
  status = internal_setent ();
  last_use = getby;

  if (status == NSS_SUCCESS)
    {
      /* Read lines until we get a definite result.  */
      do
	status = get_next_printer (name, buffer, buflen, errnop);
      while (status == NSS_RETURN);
    }

  internal_endent ();

  pthread_mutex_unlock(&lock);

  return status;
}
