/*
 * Copyright 2005 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/* Copyright (C) 2001 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Norm Jacobs <Norm.Jacobs@Sun.COM>

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#pragma ident	"$Id: nss-nis.c,v 1.1 2005/02/26 06:58:42 njacobs Exp $"

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <rpcsvc/yp.h>
#include <rpcsvc/ypclnt.h>

#include <pthread.h>

#include "nss-emulation.h"

static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;

static bool_t new_start = 1;
static char *oldkey = NULL;
static int oldkeylen = 0;


nss_status_t
_nss_nis_setprinterentry (void)
{
  pthread_mutex_lock(&lock);

  new_start = 1;
  if (oldkey != NULL)
    {
      free (oldkey);
      oldkey = NULL;
      oldkeylen = 0;
    }

  pthread_mutex_unlock(&lock);

  return NSS_SUCCESS;
}

nss_status_t
_nss_nis_endprinterentry (void)
{
  pthread_mutex_lock(&lock);

  new_start = 1;
  if (oldkey != NULL)
    {
      free (oldkey);
      oldkey = NULL;
      oldkeylen = 0;
    }

  pthread_mutex_unlock(&lock);

  return NSS_SUCCESS;
}

static nss_status_t
internal_nis_getprinterentry_r (char *buffer, size_t buflen, int *errnop)
{
  char *domain;
  char *result;
  int len;
  char *outkey;
  int keylen;
  char *p;
  nss_status_t retval;

  if (yp_get_default_domain (&domain))
    return NSS_UNAVAIL;

  if (new_start)
    retval = yperr2nss (yp_first (domain, "printers.conf.byname",
				      &outkey, &keylen, &result, &len));
  else
    retval = yperr2nss ( yp_next (domain, "printers.conf.byname", oldkey,
				      oldkeylen, &outkey, &keylen,
				      &result, &len));

  if (retval != NSS_SUCCESS) {
          if (retval == NSS_NOTFOUND)
	    *errnop = ENOENT;
	  else if (retval == NSS_TRYAGAIN)
            *errnop = errno;
          return retval;
  }

  if ((size_t) (len + 1) > buflen) {
	  free (result);
          *errnop = ERANGE;
          return NSS_TRYAGAIN;
  }
  p = strncpy (buffer, result, len);
  buffer[len] = '\0';
  while (isspace (*p))
  	++p;
  free (result);

  free (oldkey);
  oldkey = outkey;
  oldkeylen = keylen;
  new_start = 0;

  return NSS_SUCCESS;
}

nss_status_t
_nss_nis_getprinterentry_r (char *buffer, size_t buflen, int *errnop)
{
  nss_status_t status;

  pthread_mutex_lock(&lock);

  status = internal_nis_getprinterentry_r (buffer, buflen, errnop);

  pthread_mutex_unlock(&lock);

  return status;
}

nss_status_t
_nss_nis_getprinterbyname_r (const char *name, char *buffer, size_t buflen, int *errnop)
{
  nss_status_t retval;
  int parse_res;
  char *domain;
  char *result;
  int len;
  char *p;
  size_t namlen = strlen (name);

  if (name == NULL)
    {
      *errnop = EINVAL;
      return NSS_UNAVAIL;
    }

  if (yp_get_default_domain (&domain))
    return NSS_UNAVAIL;

  retval = yperr2nss (yp_match (domain, "printers.conf.byname", name, namlen,
				&result, &len));

  if (retval != NSS_SUCCESS)
    {
      if (retval == NSS_TRYAGAIN)
	*errnop = errno;
      return retval;
    }

  if ((size_t) (len + 1) > buflen)
    {
      free (result);
      *errnop = ERANGE;
      return NSS_TRYAGAIN;
    }

  p = strncpy (buffer, result, len);
  buffer[len] = '\0';
  while (isspace (*p))
    ++p;
  free (result);

  return NSS_SUCCESS;
}
